from functools import partial
from typing import Any, Callable, Dict, Tuple

from ._typing import CacheCondition


def _not_none_store_condition(result, args, kwargs, key=None) -> bool:
    return result is not None


def _store_all(result, args, kwargs, key=None) -> bool:
    return True


def _exceptions(*exceptions: Exception, default: bool = True) -> CacheCondition:
    exceptions = exceptions or Exception

    def _cond(result, args, kwargs, key):
        if isinstance(result, exceptions):
            return result
        return default

    return _cond


with_exceptions = partial(_exceptions, default=True)
only_exceptions = partial(_exceptions, default=False)


NOT_NONE = "not_none"
_ALL_CONDITIONS = {Any, "all", any, "any", None}
_NOT_NONE_CONDITIONS = {NOT_NONE, "skip_none"}


def get_cache_condition(condition: CacheCondition) -> Callable[[Any, Tuple, Dict], bool]:
    if condition in _ALL_CONDITIONS:
        return _store_all
    if condition in _NOT_NONE_CONDITIONS:
        return _not_none_store_condition
    return condition



from functools import lru_cache
from typing import Dict, Iterable, List, Match, Optional, Pattern, Tuple

from cashews._typing import TTL, Key, KeyOrTemplate, Tag, Tags, Value
from cashews.backends.interface import Backend, Callback
from cashews.exceptions import TagNotRegisteredError
from cashews.formatter import template_to_re_pattern

from .commands import CommandWrapper


class TagsRegistry:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._registry_template: Dict[Tag, List[Pattern]] = {}

    def register_tag(self, tag: Tag, key_template: KeyOrTemplate):
        self._registry_template.setdefault(tag, [])
        self._registry_template[tag].append(template_to_re_pattern(key_template))

    def get_key_tags(self, key: Key) -> Tags:
        tags = []
        for tag, patterns in self._registry_template.items():
            match = self._match_patterns(key, patterns)
            if match:
                tags.append(tag.format(**match.groupdict()))
        return tags

    @staticmethod
    def _match_patterns(key: Key, patterns: List[Pattern]) -> Optional[Match]:
        for pattern in patterns:
            match = pattern.fullmatch(key)
            if match:
                return match
        return None

    def check_tags_registered(self, key: Key, tags: Tags):
        tags = set(tags)
        key_tags = set(self.get_key_tags(key))
        difference = tags.difference(key_tags)

        if difference:
            raise TagNotRegisteredError(f"tag: {difference} not registered: call cache.register_tag before using tags")


class CommandsTagsWrapper(CommandWrapper):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self._tags_registry = TagsRegistry()
        self._tags_key_prefix = "_tag:"
        self._on_remove_cb = self._on_remove_callback()

    def setup_tags_backend(self, settings_url: str, middlewares: Tuple = (), **kwargs) -> Backend:
        return self.setup(settings_url, middlewares=middlewares, prefix=self._tags_key_prefix, **kwargs)

    @lru_cache(maxsize=1)
    def _get_tags_backend(self):
        return self._get_backend(self._tags_key_prefix)

    @property
    def tags_backend(self):
        return self._get_tags_backend()

    def register_tag(self, tag: Tag, key_template: KeyOrTemplate):
        self._tags_registry.register_tag(tag, key_template)

    def _add_backend(self, backend: Backend, *args, **kwargs):
        super()._add_backend(backend, *args, **kwargs)
        backend.on_remove_callback(self._on_remove_cb)

    def _on_remove_callback(self) -> Callback:
        async def _callback(keys: Iterable[Key], **kwargs):
            for tag, _keys in self._group_by_tags(keys).items():
                await self.tags_backend.set_remove(self._tags_key_prefix + tag, *_keys)

        return _callback

    def _group_by_tags(self, keys: Iterable[Key]) -> Dict[Tag, List[Key]]:
        tags = {}
        for key in keys:
            for tag in self.get_key_tags(key):
                tags.setdefault(tag, []).append(key)
        return tags

    def get_key_tags(self, key: Key) -> Tags:
        return self._tags_registry.get_key_tags(key)

    async def delete_tags(self, *tags: Tag):
        for tag in tags:
            await self._delete_tag(tag)

    async def _delete_tag(self, tag: Tag):
        while True:
            keys = await self.set_pop(key=self._tags_key_prefix + tag, count=100)
            if not keys:
                break
            await self.delete_many(*keys)

    async def set(
        self,
        key: Key,
        value: Value,
        expire: Optional[TTL] = None,
        exist: Optional[bool] = None,
        tags: Tags = (),
    ) -> bool:
        _set = await super().set(key=key, value=value, expire=expire, exist=exist)
        if _set and tags:
            self._tags_registry.check_tags_registered(key, tags)
            for tag in tags:
                await self.set_add(self._tags_key_prefix + tag, key, expire=expire)
        return _set

    async def incr(self, key: Key, value: int = 1, expire: Optional[float] = None, tags: Tags = ()) -> int:
        _set = await super().incr(key=key, value=value, expire=expire)
        if _set and tags:
            self._tags_registry.check_tags_registered(key, tags)
            for tag in tags:
                await self.set_add(self._tags_key_prefix + tag, key, expire=expire)
        return _set



from cashews.cache_condition import NOT_NONE  # noqa

from .bloom import bloom, dual_bloom  # noqa
from .cache.defaults import CacheDetect, context_cache_detect  # noqa
from .cache.early import early  # noqa
from .cache.fail import failover, fast_condition  # noqa
from .cache.hit import hit  # noqa
from .cache.iterator import iterator  # noqa
from .cache.simple import cache  # noqa
from .cache.soft import soft  # noqa
from .circuit_breaker import circuit_breaker  # noqa
from .locked import locked, thunder_protection  # noqa
from .rate import rate_limit  # noqa
from .rate_slide import slice_rate_limit  # noqa



import asyncio
from unittest.mock import Mock

import pytest

from cashews import Cache

pytestmark = pytest.mark.asyncio


async def test_iterator(cache: Cache):
    chunks = [b"a", b"b", b"c", b"d"]
    call = Mock()

    @cache.iterator(ttl=10, key="iterator")
    async def func():
        for chunk in chunks:
            call()
            await asyncio.sleep(0)
            yield chunk

    i = 0
    async for chunk in func():
        assert chunk == chunks[i]
        i += 1
    assert call.call_count == 4

    call.reset_mock()

    i = 0
    async for chunk in func():
        assert chunk == chunks[i]
        i += 1

    assert call.call_count == 0



import time
from functools import wraps
from typing import Optional

from cashews._typing import TTL, AsyncCallable_T, CallableCacheCondition, Decorator
from cashews.backends.interface import _BackendInterface
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from .defaults import context_cache_detect

__all__ = ("iterator",)


def iterator(
    backend: _BackendInterface,
    ttl: TTL,
    key: Optional[str] = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
) -> Decorator:
    """
    Cache decorator for iterators
    execute wrapped call and store a result with ttl if condition return true
    :param backend: cache backend
    :param ttl: duration in seconds to store a result or a callable
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    """

    ttl = ttl_to_seconds(ttl)

    def _decor(async_iterator: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(async_iterator, key=key)

        @wraps(async_iterator)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(async_iterator, _key_template, args, kwargs)

            cached = await backend.get(_cache_key)
            chunk_number = 0
            if cached:
                context_cache_detect._set(_cache_key, ttl=_ttl, name="iterator", template=_key_template)
                while True:
                    chunk = await backend.get(_cache_key + f":{chunk_number}")
                    if not chunk:
                        return
                    yield chunk
                    chunk_number += 1

            _to_cache = condition(None, args, kwargs, key=_cache_key)
            start = time.monotonic()
            async for chunk in async_iterator(*args, **kwargs):
                yield chunk
                if _to_cache:
                    await backend.set(_cache_key + f":{chunk_number}", chunk, expire=_ttl)
                chunk_number += 1
            if _to_cache:
                executing_time = time.monotonic() - start
                await backend.set(_cache_key, True, expire=_ttl - executing_time)
            return

        return _wrap

    return _decor



import logging
from functools import wraps
from typing import Any, Callable, NoReturn, Optional

from cashews._typing import TTL, AsyncCallable_T, Decorator, KeyOrTemplate
from cashews.backends.interface import _BackendInterface
from cashews.exceptions import RateLimitError
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

logger = logging.getLogger(__name__)


def _default_action(*args: Any, **kwargs: Any) -> NoReturn:
    raise RateLimitError()


def rate_limit(
    backend: _BackendInterface,
    limit: int,
    period: TTL,
    ttl: Optional[TTL] = None,
    key: Optional[KeyOrTemplate] = None,
    action: Callable = _default_action,
    prefix: str = "rate_limit",
) -> Decorator:  # pylint: disable=too-many-arguments
    """
    Rate limit for function call. Do not call function if rate limit is reached, and call given action

    :param backend: cache backend
    :param limit: number of calls
    :param period: Period
    :param ttl: time ban, default == period
    :param key: a rate-limit key template
    :param action: call when rate limit reached, default raise RateLimitError
    :param prefix: custom prefix for key, default 'rate_limit'
    """
    period = ttl_to_seconds(period)
    ttl = ttl_to_seconds(ttl) or period
    action = action or _default_action

    def decorator(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)

        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
            _period = ttl_to_seconds(period, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)

            requests_count = await backend.incr(key=_cache_key, expire=_period)
            if requests_count and requests_count > limit:
                if ttl and requests_count == limit + 1:
                    await backend.expire(key=_cache_key, timeout=_ttl)
                logger.info("Rate limit reach for %s", _cache_key)
                action(*args, **kwargs)
            return await func(*args, **kwargs)

        return wrapped_func

    return decorator



from .cache_condition import NOT_NONE, only_exceptions, with_exceptions  # noqa
from .commands import Command  # noqa
from .contrib import *  # noqa
from .decorators import context_cache_detect, fast_condition, thunder_protection  # noqa
from .exceptions import CacheBackendInteractionError, CircuitBreakerOpen, LockedError, RateLimitError  # noqa
from .formatter import default_formatter, get_template_and_func_for, get_template_for_key  # noqa
from .helpers import add_prefix, all_keys_lower, memory_limit  # noqa
from .key import get_cache_key_template, noself  # noqa
from .validation import invalidate_further  # noqa
from .wrapper import Cache, TransactionMode, register_backend  # noqa

# pylint: disable=invalid-name
cache = Cache(name="default")
noself_cache = noself(cache.cache)
failover = cache.failover
early = cache.early
soft = cache.soft
hit = cache.hit
transaction = cache.transaction
setup = cache.setup
cache_detect = cache.detect

circuit_breaker = cache.circuit_breaker
dynamic = cache.dynamic
rate_limit = cache.rate_limit
slice_rate_limit = cache.slice_rate_limit
locked = cache.locked

invalidate = cache.invalidate
invalidate_func = cache.invalidate_func


mem = Cache(name="mem")
mem.setup(
    "mem://?check_interval=1",
    size=1_000_000,
)  # 1_000_000 * 248(small dict size) == 31 mb
